#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef HAVE_LIBPROTEINGEOMETRY
#include <proteingeometry/dist.h>
#endif

#include "pgc.h"
#include "pgc_binding.h"
#include "pgc_conf.h"
#include "pgc_conf_lexer.h"
#include "pgc_evaluate.h"
#include "pgc_function.h"
#include "pgc_hash.h"
#include "pgc_record.h"

#define ERROR_MESSAGE "Command-line argument format: \n\
  pgc [source.pgc] [-c (or --config) optional.conf] [-h (or --help)]\n" 

#define INTRO_MESSAGE "Protein Geometry Calculator v1.0b,\n\
by Julian Graham (julian.graham@yale.edu)\n"

struct pgc_conf_entry **conf_table;

/* A function for printing the PGC data types -- everything except the vector
   type, since that is handled on its own... */
void print_data (int data_type, void *data) {
  switch (data_type) {
  case PGC_NIL:
    printf ("(nil)");
    break;
  case PGC_INT:
    printf ("%d", *((int *) data));
    break;
  case PGC_FLOAT:
    printf ("%lf", *((double *) data));
    break;
  case PGC_STRING:
    printf ("\"%s\"", (char *) data);
    break;
  case PGC_RECORD:
    if (data == NULL)
      printf ("(record) nil");
    else
      printf ("(record of type \"%s\")", 
              ((struct pgc_record *) data)->record_type_name);
    break;
  default:
    break;
  }
  return;
}

void inputloop (char *expression, 
		struct pgc_environment *initial_environment) {
  int i = 0;
  struct pgc_environment *new_environment = initial_environment;
  do {
    new_environment = evaluate ("stdin", expression, new_environment);
    printf ("result = ");
    switch (new_environment->result_type) {
    case PGC_STRING:
      printf ("(string) ");
      print_data (new_environment->result_type, new_environment->result);
      printf ("\n");
      break;
    case PGC_NIL:
      print_data (new_environment->result_type, new_environment->result);
      printf ("\n");
      break;
    case PGC_INT:
      printf ("(int) ");
      print_data (new_environment->result_type, new_environment->result);
      printf ("\n");
      break;
    case PGC_FLOAT:
      printf ("(float) ");
      print_data (new_environment->result_type, new_environment->result);
      printf ("\n");
      break;
    case PGC_RECORD:
      print_data (new_environment->result_type, new_environment->result);
      printf ("\n");
      break;
    case PGC_VECTOR:
      printf ("(vector) [");
      switch (((struct pgc_vector *) new_environment->result)->vector_type) {
      case PGC_VECTOR:
	printf 
	  ("(%d subvector",
	   ((struct pgc_vector *) new_environment->result)->vector_length);
        if (((struct pgc_vector *) new_environment->result)->vector_length > 1)
          printf ("s");
        printf (")");
	break;
      case PGC_NIL:
	printf 
	  ("(nil * %d)",
	   ((struct pgc_vector *) new_environment->result)->vector_length);
	break;
      case PGC_RECORD:
        printf
          ("(record * %d)",
           ((struct pgc_vector *) new_environment->result)->vector_length);
        break;
      default:
        for (i = 0; 
	     i < ((struct pgc_vector *) new_environment->result)->
	       vector_length;
	     i++) {
          print_data 
            (((struct pgc_vector *) new_environment->result)->vector_type,
             ((void **) ((struct pgc_vector *) new_environment->result)->
                vector)[i]);
          if (i != ((struct pgc_vector *) new_environment->result)->
                vector_length - 1)
            printf (", "); 
        }        
	break;
      }
      printf ("]\n");
      break;
    default:
      printf ("(undefined type)\n");
      break;
    }
    printf (">> ");
    bzero (expression, strlen (expression));
    expression = fgets (expression, PGC_MAX_EXPR_LENGTH, stdin);
    if (expression != NULL)
      expression[(strlen (expression)) - 1] = (char) 0;
  }
  while (expression != NULL);
}

void pgc_init_conf_table (char *filename) {
  extern int yy_conflineno;
  extern int yy_conflex ();
  extern struct pgc_conf_entry yylval;
  extern void *yy_confrestart (FILE *file_ptr);

  conf_table = calloc (PGC_CONF_TABLE_SIZE, sizeof (struct pgc_conf_entry *));
  FILE *conf_ptr = fopen (filename, "r");
  if (conf_ptr != NULL) {
    yy_confrestart (conf_ptr);
    int break_flag = 0;
    printf ("Reading configuration data from \'%s\'... ", filename); 
    do {
      int lex_return = yy_conflex ();
      switch (lex_return) {
      case PGC_CONF_EOF:
	printf ("done.\n");
	break_flag = 1;
	break;
      case PGC_CONF_TOKEN:
	pgc_add_conf_entry (yylval.entry_name, yylval.entry_value);
	break;
      case PGC_CONF_ERROR:
	printf ("\nError in configuration file at line %d.\n",
		yy_conflineno);
	break_flag = 1;
      default:
	break;
      }
      if (break_flag)
	break;
    }
    while (1);
    fclose (conf_ptr);
  }
  else printf ("Configuration file \'%s\' missing or unreadable.\n", filename);
  return;
}

int main (int argc, char *argv[]) {
  struct pgc_environment *environment;
  char *expression = NULL;
  int current_max_expression_length = 0;
  char *config_file = strdup ("pgc.conf");
  char **source_files = NULL;
  int num_source_files = 0;

  environment = pgc_copy_environment (NULL);

  /* yy_flex_debug = 0; */

  #ifdef HAVE_LIBPROTEINGEOMETRY
  default_initialization (); /* Initialize the protein geometry library... */
  #endif

  pgc_init_function_table ();
  pgc_init_record_table ();

  expression = strdup ("");

  int i = 1;
  while (i < argc) {
    if ((strcmp (argv[i], "-c") == 0) || (strcmp (argv[i], "--config") == 0)) {
      if ((i + 1) < argc) {
	free (config_file);
	config_file = strdup (argv[i + 1]);
	i++;
      }
      else {
	printf (ERROR_MESSAGE);
	return 1;
      }
      i++;
      continue;
    }
    if ((strcmp (argv[i], "-h") == 0) || (strcmp (argv[i], "--help") == 0)) {
      printf (ERROR_MESSAGE);
      return 0;
    }
    else {
      source_files = realloc (source_files, sizeof (char *) * 
			      (num_source_files + 1));
      source_files[num_source_files] = strdup (argv[i]);
      num_source_files++;
      i++;
      continue;
    }
  }
  
  /* (void) umask (0600); */
  pgc_init_conf_table (config_file);
  printf (INTRO_MESSAGE);

  for (i = 0; i < num_source_files; i++) {
    FILE *file_input = fopen (source_files[i], "r");
    if (file_input == NULL)
      fprintf (stderr, "Unable to open file '%s' for reading.\n", argv[i]);
    else {
      char c;
      while (!feof (file_input)) {
	(void) fread (&c, 1, 1, file_input);
	if (strlen (expression) == current_max_expression_length) {
	  expression = 
	    realloc (expression, 
		     (sizeof (char) * (current_max_expression_length + 2)));
	  expression[current_max_expression_length] = c;
	  expression[current_max_expression_length + 1] = (char) NULL;
	  current_max_expression_length += 1;
	}
	else {
	  int insertion_pos = strlen (expression);
	  expression[insertion_pos] = c;
	  expression[insertion_pos + 1] = (char) NULL;
	}
      }
      fclose (file_input);
      environment = evaluate (source_files[i], expression, environment);
      if (strlen (expression) > 0)
	bzero (expression, strlen (expression));
    }
  }

  if (current_max_expression_length < PGC_MAX_EXPR_LENGTH)
    expression = realloc (expression, sizeof (char) * PGC_MAX_EXPR_LENGTH);
  inputloop (expression, environment);
  return 0;
}
